#include <stdio.h>
#include <string.h>

#include "prompt.h"
#include "cryptsetup.h"
#include "tpm.h"

static void
prompt_existing_password (Buffer *password)
{  
    prompt ("Enter password", password);
}

static TPM_BOOL
prompt_new_password (Buffer *new_password)
{  
    prompt ("Enter new password", new_password);
    Buffer retyped_password;
    prompt ("Retype password", &retyped_password);
    TPM_BOOL matches = buffer_matches (*new_password, retyped_password);
    buffer_free (retyped_password);
    if (!matches)
    {
        buffer_free (*new_password);
        fprintf (stderr, "Passwords do not match\n");
        return FALSE;
    }
  
    return TRUE;
}

static TPM_BOOL
get_dek (Buffer *dek, UINT32 index)
{
    Buffer password;
    prompt_existing_password (&password);

    /* Get DEK from TPM */
    if (!tpm_read_nvram (index, password, 32, dek))
        return FALSE;

    return TRUE;
}

static int
format (const char *device, UINT32 index)
{
    // FIXME: Check if this device exists and is *not* setup to use cryptsetup-tpm

    /* Prompt for a password to encrypt it with */
    Buffer password;
    if (!prompt_new_password (&password))
        return FALSE;

    /* Create a new DEK */
    // FIXME
    Buffer dek;
    buffer_init_random (32, &dek);

    if (!cryptsetup_format (device, dek))
        return FALSE;

    /* Write DEK to NVRAM */
    if (!tpm_create_nvram (index, password, 32) ||
        !tpm_write_nvram (index, password, dek))
        return EXIT_FAILURE;

    buffer_free (password);
    buffer_free (dek);

    return EXIT_SUCCESS;
}

static int
add_key (const char *device, UINT32 index)
{
    // FIXME: Check if this device exists and is *not* setup to use cryptsetup-tpm

    Buffer old_password;
    prompt_existing_password (&old_password);

    Buffer password;
    if (!prompt_new_password (&password))
        return EXIT_FAILURE;

    /* Create a new DEK */
    // FIXME
    Buffer dek;
    buffer_init_random (32, &dek);

    /* Update LUKS password */
    if (!cryptsetup_add_key (device, old_password, dek))
        return EXIT_FAILURE;
    buffer_free (old_password);

    // FIXME: Revert cryptsetup change if fail to write to TPM

    if (!tpm_create_nvram (index, password, 32))
        return EXIT_FAILURE;

    /* Write DEK to NVRAM */
    if (!tpm_write_nvram (index, password, dek))
        return EXIT_FAILURE;

    buffer_free (password);
    buffer_free (dek);

    return EXIT_SUCCESS;
}

static int
change_key (const char *device, UINT32 index)
{
    // FIXME: Check if this device exists and is setup to use cryptsetup-tpm

    Buffer old_dek;
    if (!get_dek (&old_dek, index))
        return EXIT_FAILURE;

    Buffer password;
    if (!prompt_new_password (&password))
        return EXIT_FAILURE;

    /* Create a new DEK */
    // FIXME
    Buffer dek;
    buffer_init_random (32, &dek);

    /* Update LUKS password */
    if (!cryptsetup_change_key (device, old_dek, dek))
        return EXIT_FAILURE;
    buffer_free (old_dek);

    // FIXME: Revert cryptsetup change if fail to write to TPM

    /* Update DEK in NVRAM */
    if (!tpm_write_nvram (index, password, dek))
        return EXIT_FAILURE;

    return EXIT_SUCCESS;
}

int
main (int argc, char **argv)
{
    if (argc < 2)
    {
        fprintf (stderr, "You must specify an action. See '%s --help' for details\n", argv[0]);
        return EXIT_FAILURE;
    }

    char *action_args[argc];
    int action_args_length = 0;
    UINT32 index = 0x60000000; // FIXME: hard-coded

    /* Process options */
    int i;
    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] != '-')
        {
            action_args[action_args_length] = argv[i];
            action_args_length++;
            continue;
        }
            
        if (strcmp (argv[i], "--help") == 0)
        {
            fprintf (stderr, "Usage: %s [OPTION...] <action> <action-specific>\n"
                             "      -i, --index     TPM NVRAM index to use\n"
                             "      --version       Show release version\n"
                             "      --help          Show this help message\n"
                             "\n"
                             "<action> is one of:\n"
                             "        luksFormat <device> - formats a LUKS device with a TPM key\n"
                             "        luksAddKey <device> - add a TPM key to a LUKS device\n"
                             "        luksChangeKey <device> - change the TPM key on a LUKS device\n"
                             "\n"
                             "<device> is the encrypted device\n", argv[0]);
            return EXIT_SUCCESS;
        }
        else if (strcmp (argv[i], "-i") == 0 || strcmp (argv[i], "--index") == 0)
        {
            if (i + 1 >= argc)
            {
                fprintf (stderr, "Option %s requires an argument\n", argv[i]);
                return EXIT_FAILURE;
            }   
            if (strncmp (argv[i+1], "0x", 2) == 0)
                index = strtol (argv[i+1] + 2, NULL, 16);
            else
                index = strtol (argv[i+1], NULL, 10);
            i++;
        }
        else if (strcmp (argv[i], "--version") == 0)
        {
            fprintf (stderr, "%s %s\n", argv[0], VERSION);
            return EXIT_SUCCESS;
        }
        else
        {
            fprintf (stderr, "Unknown option '%s'. See '%s --help' for details\n", argv[1], argv[0]);
            return EXIT_FAILURE;
        }
    }
  
    if (action_args_length == 0)
    {
        fprintf (stderr, "Missing action. See '%s --help' for details\n", argv[0]);
        return EXIT_FAILURE;
    }

    /* Process actions */
    if (strcmp (action_args[0], "luksFormat") == 0)
    {
        if (action_args_length != 2)
        {
            fprintf (stderr, "luksFormatKey requires <device> as arguments. See '%s --help' for details\n", argv[0]);
            return EXIT_FAILURE;
        }
        return format (action_args[1], index);
    }
    else if (strcmp (action_args[0], "luksAddKey") == 0)
    {
        if (action_args_length != 2)
        {
            fprintf (stderr, "luksAddKey requires <device> as arguments. See '%s --help' for details\n", argv[0]);
            return EXIT_FAILURE;
        }
        return add_key (action_args[1], index);
    }
    else if (strcmp (action_args[0], "luksChangeKey") == 0)
    {
        if (action_args_length != 2)
        {
            fprintf (stderr, "luksChangeKey requires <device> as arguments. See '%s --help' for details\n", argv[0]);
            return EXIT_FAILURE;
        }
        return change_key (action_args[1], index);
    }
    else
    {
        fprintf (stderr, "Unknown action '%s'. See '%s --help' for details\n", action_args[0], argv[0]);
        return EXIT_FAILURE;
    }
}
