#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "cryptsetup.h"

static TPM_BOOL
run_cryptsetup (char *const argv[], pid_t *pid, int *stdin_fd)
{
    /* Run cryptsetup with a pipe to stdin */
    int p[2];
    if (pipe (p) < 0)
    {
        fprintf (stderr, "Failed to make pipe: %s\n", strerror (errno));
        return FALSE;
    }

    *pid = fork ();
    if (*pid < 0)
    {
        fprintf (stderr, "Failed to fork: %s\n", strerror (errno));
        close (p[0]);
        close (p[1]);
        return FALSE;
    }
    if (*pid == 0)
    {
        dup2 (p[0], STDIN_FILENO);
        close (p[0]);
        close (p[1]);
        execvp ("cryptsetup", argv);
        _exit (EXIT_FAILURE);
    }
    close (p[0]);
    *stdin_fd = p[1];
  
    return TRUE;
}

static TPM_BOOL
wait_cryptsetup (pid_t pid)
{
    /* Wait to confirm this worked */
    int status;
    if (waitpid (pid, &status, 0) < 0)
    {
        fprintf (stderr, "Failed to run cryptsetup: %s\n", strerror (errno));
        return FALSE;
    }
    if (WIFEXITED (status))
    {
        if (WEXITSTATUS (status) != EXIT_SUCCESS)
        {
            fprintf (stderr, "cryptsetup exited with status %d\n", WEXITSTATUS (status));
            return FALSE;
        }
    }
    else if (WIFSIGNALED (status))
    {
        fprintf (stderr, "cryptsetup exited with signal %d\n", WTERMSIG (status));
        return FALSE;
    }
    else
    {
        fprintf (stderr, "Failed to run cryptsetup: %s\n", strerror (errno));
        return FALSE;
    }

    return TRUE;
}

TPM_BOOL
cryptsetup_format (const char *device, Buffer key)
{
    char *argv[5];
    argv[0] = "cryptsetup";
    argv[1] = "luksFormat";
    argv[2] = (char *) device;
    argv[3] = "-";
    argv[4] = NULL;
  
    pid_t cryptsetup_pid;
    int stdin_fd;
    if (!run_cryptsetup (argv, &cryptsetup_pid, &stdin_fd))
        return FALSE;
 
    /* Write the key to cryptsetup */
    if (write (stdin_fd, key.data, key.length) != key.length)
    {
        fprintf (stderr, "Failed to write key to cryptsetup: %s\n", strerror (errno));
        kill (cryptsetup_pid, SIGTERM);
        close (stdin_fd);
        return FALSE;
    }
    close (stdin_fd);

    return wait_cryptsetup (cryptsetup_pid);
}

static TPM_BOOL
add_or_change_key (const char *device, const char *action, Buffer old_key, Buffer new_key)
{
    char *argv[9];
    argv[0] = "cryptsetup";
    argv[1] = (char *) action,
    argv[2] = (char *) device;
    argv[3] = "--key-file";
    argv[4] = "-";
    argv[5] = "--keyfile-size";
    char old_length[1024];
    snprintf (old_length, 1024, "%zi", old_key.length);
    argv[6] = old_length;
    argv[7] = "-";
    argv[8] = NULL;
  
    pid_t cryptsetup_pid;
    int stdin_fd;
    if (!run_cryptsetup (argv, &cryptsetup_pid, &stdin_fd))
        return FALSE;
 
    /* Write the old and new keys to cryptsetup */
    if (write (stdin_fd, old_key.data, old_key.length) != old_key.length ||
        write (stdin_fd, new_key.data, new_key.length) != new_key.length)
    {
        fprintf (stderr, "Failed to write key to cryptsetup: %s\n", strerror (errno));
        kill (cryptsetup_pid, SIGTERM);
        close (stdin_fd);
        return FALSE;
    }
    close (stdin_fd);

    return wait_cryptsetup (cryptsetup_pid);
}

TPM_BOOL
cryptsetup_add_key (const char *device, Buffer old_key, Buffer new_key)
{
    return add_or_change_key (device, "luksAddKey", old_key, new_key);
}

TPM_BOOL
cryptsetup_change_key (const char *device, Buffer old_key, Buffer new_key)
{
    return add_or_change_key (device, "luksChangeKey", old_key, new_key);
}
