#include <stdio.h>
#include <trousers/tss.h>
#include <trousers/trousers.h>

#include "tpm.h"

static TPM_BOOL
connect_to_tpm (TSS_HCONTEXT *context)
{
    TSS_RESULT result = Tspi_Context_Create (context);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to create Tspi context: %s\n", Trspi_Error_String (result));
        return FALSE;
    }
    result = Tspi_Context_Connect (*context, NULL);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to connect Tspi context: %s\n", Trspi_Error_String (result));
        return FALSE;
    }

    return TRUE;
}

static TPM_BOOL
set_tpm_secret (TSS_HCONTEXT context, TSS_FLAG secret_mode, BYTE *secret, UINT32 secret_length)
{
    TSS_HTPM tpm;
    TSS_HPOLICY policy;
    TSS_RESULT result = Tspi_Context_GetTpmObject (context, &tpm);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to get TPM object: %s\n", Trspi_Error_String (result));
        return FALSE;
    }  
    result = Tspi_GetPolicyObject (tpm, TSS_POLICY_USAGE, &policy);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to get policy: %s\n", Trspi_Error_String (result));
        return FALSE;
    }  
    result = Tspi_Policy_SetSecret (policy, secret_mode, secret_length, secret);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to set secret: %s\n", Trspi_Error_String (result));
        return FALSE;
    }  

    return TRUE;
}

static TPM_BOOL
create_nvram_object (TSS_HCONTEXT context, UINT32 index, TSS_HNVSTORE *object)
{
    TSS_RESULT result = Tspi_Context_CreateObject (context, TSS_OBJECT_TYPE_NV, 0, object);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to create NVRAM object: %s\n", Trspi_Error_String (result));
        return FALSE;
    }
    result = Tspi_SetAttribUint32 (*object, TSS_TSPATTRIB_NV_INDEX, 0, index);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to set NVRAM index: %s\n", Trspi_Error_String (result));
        return FALSE;
    }

    return TRUE;
}

static TPM_BOOL
set_nvram_data_size (TSS_HNVSTORE object, UINT32 data_size)
{
    TSS_RESULT result = Tspi_SetAttribUint32 (object, TSS_TSPATTRIB_NV_DATASIZE, 0, data_size);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to set NVRAM size: %s\n", Trspi_Error_String (result));
        return FALSE;
    }

    return TRUE;
}

static TPM_BOOL
set_nvram_permissions (TSS_HNVSTORE object, UINT32 permissions)
{
    TSS_RESULT result = Tspi_SetAttribUint32 (object, TSS_TSPATTRIB_NV_PERMISSIONS, 0, permissions);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to set NVRAM permissions: %s\n", Trspi_Error_String (result));
        return FALSE;
    }

    return TRUE;
}

static TPM_BOOL
assign_usage_policy (TSS_HCONTEXT context, TSS_HOBJECT object, TSS_FLAG secret_mode, BYTE *secret, UINT32 secret_length)
{
    TSS_HPOLICY policy;
    TSS_RESULT result = Tspi_Context_CreateObject (context, TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &policy);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to create usage policy: %s\n", Trspi_Error_String (result));
        return FALSE;
    }
    result = Tspi_Policy_SetSecret (policy, secret_mode, secret_length, secret);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to set secret: %s\n", Trspi_Error_String (result));
        return FALSE;
    }   
    result = Tspi_Policy_AssignToObject (policy, object);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to assign policy: %s\n", Trspi_Error_String (result));
        return FALSE;
    }
   
    return TRUE;
}

TPM_BOOL
tpm_create_nvram (UINT32 index, Buffer password, UINT32 length)
{
    TSS_HCONTEXT tsp_context;
    BYTE tpm_password[TCPA_SHA1_160_HASH_LEN] = TSS_WELL_KNOWN_SECRET;
    TSS_HNVSTORE nvram_object;
    if (!connect_to_tpm (&tsp_context) ||
        !set_tpm_secret (tsp_context,  TSS_SECRET_MODE_SHA1, tpm_password, TCPA_SHA1_160_HASH_LEN) ||
        !create_nvram_object (tsp_context, index, &nvram_object) ||
        !set_nvram_data_size (nvram_object, length) ||
        !set_nvram_permissions (nvram_object, TPM_NV_PER_AUTHREAD | TPM_NV_PER_AUTHWRITE) ||
        !assign_usage_policy (tsp_context, nvram_object, TSS_SECRET_MODE_PLAIN, password.data, password.length))
        return FALSE;

    TSS_RESULT result = Tspi_NV_DefineSpace (nvram_object, 0, 0);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to create NVRAM area: %s\n", Trspi_Error_String (result));
        return FALSE;
    }

    return TRUE;
}

TPM_BOOL
tpm_read_nvram (UINT32 index, Buffer password, UINT32 length, Buffer *result)
{
    TSS_HCONTEXT tsp_context;
    TSS_HNVSTORE nvram_object;
    if (!connect_to_tpm (&tsp_context) ||
        !create_nvram_object (tsp_context, index, &nvram_object) ||
        !assign_usage_policy (tsp_context, nvram_object, TSS_SECRET_MODE_PLAIN, password.data, password.length))
        return FALSE;

    buffer_init (length, result);
    UINT32 l = length;
    TSS_RESULT r = Tspi_NV_ReadValue (nvram_object, 0, &l, &result->data);
    if (r != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to read NVRAM value: %s\n", Trspi_Error_String (r));
        return FALSE;
    }
    result->length = l;
  
    return TRUE;
}

TPM_BOOL
tpm_write_nvram (UINT32 index, Buffer password, Buffer value)
{
    TSS_HCONTEXT tsp_context;
    TSS_HNVSTORE nvram_object;
    if (!connect_to_tpm (&tsp_context) ||
        !create_nvram_object (tsp_context, index, &nvram_object) ||
        !assign_usage_policy (tsp_context, nvram_object, TSS_SECRET_MODE_PLAIN, password.data, password.length))
        return FALSE;

    TSS_RESULT result = Tspi_NV_WriteValue (nvram_object, 0, value.length, value.data);
    if (result != TSS_SUCCESS)
    {
        fprintf (stderr, "Failed to write NVRAM value: %s\n", Trspi_Error_String (result));
        return FALSE;
    }
  
    return TRUE;
}
